import abc
from enum import Enum
from pydantic import BaseModel


class W24SizeType(str, Enum):
    NOMINAL = "NOMINAL"
    DIAMETER = "DIAMETER"
    WIDTHS_ACCROSS_FLATS = "WIDTHS_ACCROSS_FLATS"


class W24Size(BaseModel, abc.ABC):
    """ Abstract Base Class for the Sizes
    """

    blurb: str
    """ blurb for human consumption
    """

    size_type: W24SizeType
    """ size type for deserialization
    """

    nominal_size: float
    """ Nominal size
    """


class W24SizeNominal(W24Size):
    """ Exactly your nominal size
    """
    size_type = W24SizeType.NOMINAL


class W24SizeDiameter(W24Size):
    """ Diameter size
    """
    size_type = W24SizeType.DIAMETER


class W24SizeWidthsAcrossFlats(W24Size):
    """ Widths accross flats / Wrench Sizes
    """
    size_type = W24SizeType.WIDTHS_ACCROSS_FLATS

    width_accross_flats: float
    """ Size accross flats or Wrench sizes
    """



import abc
from enum import Enum
from typing import List, Optional, Tuple

from pydantic import BaseModel, UUID4


class W24GDTCharacteristic(str, Enum):
    """ Enum of all possible Characteristics
    following ISO 1101.

    NOTE: when the GDT frame only describes the location
    of a reference datum, the value DATUM_INDICATOR is used
    """
    FORM_STRAIGHTNESS = "⏤"
    FORM_FLATNESS = "⏥"
    FORM_CIRCULARITY = "○"
    FORM_CYLINDRICITY = "⌭"
    PROFILE_OF_SURFACE = "⌓"
    PROFILE_OF_LINE = "⌒"
    ORIENTATION_PERPENDICULARITY = "⟂"
    ORIENTATION_ANGULARITY = "∠"
    ORIENTATION_PARALLELISM = "∥"
    LOCATION_POSITION = "⌖"
    LOCATION_CONCENTRICITY = "◎"
    LOCATION_SYMMETRY = "⌯"
    RUNOUT_CIRCULAR = "↗"
    RUNOUT_TOTAL = "⌰"
    DATUM_INDICATOR = "[DATUM]"


class W24GDTFeatureAssociated(str, Enum):
    """ Enum of all associated toleranced features
    """
    MINIMAX = "Ⓒ"
    GAUSSIAN = "Ⓖ"
    MIN_CIRCUMSCRIBED = "Ⓝ"
    MAX_CIRCUMSCRIBED = "Ⓧ"
    TANGENT = "Ⓣ"


class W24GDTFeatureDerived(str, Enum):
    """ Enum of all derived features
    """
    PROJECTED = "Ⓟ"
    MEAN = "Ⓐ"


class W24GDTZoneCombination(str, Enum):
    """ Enum of all tolerance zonce combinations
    """
    COMBINED = "CZ"
    SEPARATED = "SZ"


class W24GDTZoneShape(str, Enum):
    """ Enum of all zone shapes
    """
    DIAMETER = "Ø"
    DIAMETER_SPHERE = "S⌀"


class W24GDTZoneConstraint(str, Enum):
    """ Enum of the Zone Contraints
    """
    UNSPECIFIED_INCLINATION = "OZ"
    UNSPECIFIED_OFFSET = "VA"


class W24GDTDatum(BaseModel):
    """ Preliminary implementation of the GD&T Datum

    NOTE: future implementations might allow fine-grained
    access to the attributes of complex names:
    e.g., (A-B-C-D)[CM]
    """

    blurb: str
    """ Reference name. Typically: A,B,C ...
    Can also contain more complex names, e.g., (A-B-C-D)[CM]
    """


class W24GDTZoneOffset(BaseModel):
    """ Specified offset indicated
    by leading UZ...
    """

    blurb: str
    """ blurb of the specified offset for human consumption.
    e.g., UZ+0.15, UZ-0.2, UZ+0.1:0.2
    """

    # offset_min: float
    """ signed offset with the smaller absolute value
    e.g., -0.1
    """

    # offset_max: float
    """ signed offset with the larger absolute value
    e.g., -0.3
    """


class W24GDTZoneValue(BaseModel):
    """ Preliminary defintion of the GDT Zone Value
    Future implementation will give access to the
    width and extend seperately
    """

    blurb: str
    """ String representation for human consumption
    e.g., 0.05/12x10°
    """

    width_min: float
    """ Minimal width """

    width_max: Optional[float]
    """ Optional maximal width """

    extend_quantity: Optional[int]
    """ Optional quantity of the spacing """

    extend_shape: Optional[W24GDTZoneShape]
    """ Optional shape of the extend """

    extend: Optional[float]
    """ Optional extend """

    extend_angle: Optional[float]
    """ Optional angle of the extend spacing """


class W24GDTFilterType(str, Enum):
    """ Preliminary list of feature filters

    Filters remaining: RG, S, OH, SW, AB, CW
    """
    GAUSSIAN = "G"


class W24GDTFilter(BaseModel, abc.ABC):
    """ Abstract base class to describe feature filters
    """

    blurb: str
    """ String representation of the file for human consumption
    """

    filter_type: W24GDTFilterType
    """ Filter Type to facilitate deserialization
    """


# class W24GDTFilterG(W24GDTFilter):
#     """ Gaussian Feature filter without contraints
#     """
#     filter_type = W24GDTFilterType.GAUSSIAN

#     wavelength_min: float
#     """ Minimum wavelength in wavers/revolution
#     """

#     wavelength_max: float
#     """ Maximum wavelength in wavers/revolution
#     """


class W24GDTReferenceAssociation(str, Enum):
    """ Association of the Reference Element
    """

    MINIMAX = "C"
    """ Minimax (Tschebyschew-) Element without contraint
    """

    MINIMAX_EXTERNAL = "CE"
    """ Minimax (Tschebyschew-) Element with external
    constraint
    """

    MINIMAX_INTERNAL = "CI"
    """ Minimal (Tschebyschew-) Element with internal
    constraint
    """

    GAUSSIAN = "G"
    """ Gaussian Least Square Element without constraint
    """

    GAUSSIAN_EXTERNAL = "GE"
    """ Gaussian Least Square Element with external
    constraint
    """

    GAUSSIAN_INTERNAL = "GI"
    """ Gaussian Least Square Element with external
    constraint
    """

    MIN_CIRCUMSCRIBED = "N"
    """ Minimal circumscribed element
    """

    MAX_CIRCUMSCRIBED = "X"
    """ Maximal circumscribed element
    """


class W24GDTReferenceParameter(str, Enum):
    """ Parameter of the reference element
    """

    PEAK_VALUE = "P"
    """ Highest value
    """

    VALLEY_VALUE = "V"
    """ Lowest value
    """

    DEVISATION_SPAN = "T"
    """ Deviation span
    """

    STANDARD_DEVIATION = "Q"
    """ Root mean square deviation
    """


class W24GDTMaterialCondition(str, Enum):
    """ Enum for Material Conditions
    """

    MAXIMUM = "Ⓜ"
    """ Maximum material condition (MMC)
    """

    MINIMUM = "Ⓛ"
    """ Mimimal material condition (LMC)
    """

    RECIPROCITY = "Ⓡ"
    """ Reciprocity
    """


class W24GDTState(str, Enum):
    """ Enum for the State
    """
    FREE = "Ⓕ"


class W24GDTFrame(BaseModel):
    """ Representation of the Geometric Dimensioning
    and Toleration frame
    """
    gdt_id: Optional[UUID4] = None
    """ Unique id of the GDT
    """

    blurb: str
    """ String representation of the label for human consumption
    e.g., [⌖|⌀0.3Ⓜ|A|B|C]
    """
    characteristic: W24GDTCharacteristic
    """ Section for gemetric characteristic e.g.: ⌓
    """

    zone_shape: Optional[W24GDTZoneShape] = None
    """ Tolerance zone: shape
    e.g, S⌀
    """

    zone_value: Optional[W24GDTZoneValue]
    """ GDT value: e.g., 0.03
    Need to be optional to supprt Datum Feature Indicators
    """

    zone_combinations: List[W24GDTZoneCombination] = []
    """ Ordered list of zone combinations, e.g., CZ, SZ
    """

    zone_offset: Optional[W24GDTZoneOffset] = None
    """ Optional specified offset, e.g., UZ-0.2
    """

    zone_constraint: Optional[W24GDTZoneConstraint] = None
    """ Optional zone constraint: e.g., OZ, VA
    """

    feature_filter: Optional[W24GDTFilter] = None
    """ Optional feature filter
    """

    feature_associated: Optional[W24GDTFeatureAssociated] = None
    """ Associated toleraced feature
    """

    feature_derived: Optional[W24GDTFeatureDerived] = None
    """ Derived Feature
    """

    reference_association: Optional[W24GDTReferenceAssociation] = None
    """ Reference element association
    """

    reference_parameter: Optional[W24GDTReferenceParameter] = None
    """ Reference element parameter
    """

    material_condition: Optional[W24GDTMaterialCondition] = None
    """ Material condition
    """

    state: Optional[W24GDTState] = None
    """ State: FREE or None
    """

    data: List[W24GDTDatum] = []
    """ Ordered list of data
    """


class W24GDT(BaseModel):
    """ Parent object for Geometric Dimensionsing and Toleration
    Frames, attaching them to the physical location on the drawing.
    """

    bounding_polygon: List[Tuple[float, float]]
    """ bounding polygon of the GDT annotation as tuple of x, y
    coordinates in the pixel coordinate system of the sectional.

    NOTE: simple GDTs are represented by a simple rectangle.
    To support more complex GDTs (e.g., with a Sectional
    Plane Indicator, we define a polygon)

    NOTE: the polygon starts at the top left and is
    oriented clock-wise.
    """

    frame: W24GDTFrame
    """ Representation of the GDT frame
    """

    # measure_label: Optional[W24MeasureLabel] = None
    """ Optional size dimension, typically annotated
    above the feature control frame.

    Example:
        ⌀12.0 + /- 0.1
        [⊥|⌀0.1Ⓜ| A]
    """

    # sectional_plane
    """ Definition of Sectional Plane Indicator will
    follow.
    Example: ⟨[∥| A]
    """  # pylint: disable=pointless-string-statement

    # orientation_plane
    """ Definition of the Orientational Plane Indicator
    will follow
    Example: ⟨[∥| A]⟩
    """  # pylint: disable=pointless-string-statement

    # direction_feature
    """ Defintion of the Direction Feature Indicator
    will follow
    Example: ←[∥| A]
    """  # pylint: disable=pointless-string-statement

    # collection_plane
    """ Defintion of the Collection Plane Indicator
    will follow
    Exmple: ◯[∥| B]
    """  # pylint: disable=pointless-string-statement



from enum import Enum


class FileFormatCad(str, Enum):
    """ List of the supported CAD file formats

    NOTE: at this stage, the API only supports DXF
    as output format. Additional formats are available
    upon request.
    """
    DXF = "DXF"


class FileFormatImage(str, Enum):
    """ List of supported File Formats in which
    the Thumbnail can be supplied.

    NOTE: At this stage, the API only supports JPEG.
    This prepares the support for additional format
    (e.g. TIFF)
    """
    JPEG = "JPEG"



""" W24Client Module

DESCRIPTION
    The module contains everything that is needed to
    communicate with the W24 API - allowing you
    to interpret the contents of your technical drawings.

AUTHOR
    Jochen Mattes (Werk24)

EXAMPLE
    # obtain the thumbnail of a page
    drawing_bytes = open(...,"r").read()
    client = W24TechreadClient.make_from_env()
    await client.read_drawing_with_hooks(
        drawing_bytes,
        [Hook(
                ask=W24AskPageThumbnail(),
                callback=lambda msg: print("Received Thumbnail of Page")
        ]))
"""
import asyncio
import logging
import os
from types import TracebackType
from typing import AsyncGenerator, Callable, List, Optional, Type

import dotenv
from pydantic import BaseModel

from werk24.auth_client import AuthClient
from werk24.exceptions import (LicenseError, RequestTooLargeException,
                               ServerException)
from werk24.models.ask import W24Ask
from werk24.models.techread import (W24TechreadAction,
                                    W24TechreadInitResponse,
                                    W24TechreadMessage,
                                    W24TechreadMessageSubtype,
                                    W24TechreadMessageType, W24TechreadRequest)
from werk24.techread_client_https import TechreadClientHttps
from werk24.techread_client_wss import TechreadClientWss

# make the logger
logger = logging.getLogger(  # pylint: disable=invalid-name
    'w24_techread_client')

ENVIRONS = [
    "W24TECHREAD_SERVER_HTTPS",
    "W24TECHREAD_SERVER_WSS",
    "W24TECHREAD_VERSION",
    "W24TECHREAD_AUTH_CLIENT_ID",
    "W24TECHREAD_AUTH_CLIENT_SECRET",
    "W24TECHREAD_AUTH_IDENTITY_POOL_ID",
    "W24TECHREAD_AUTH_USER_POOL_ID",
    "W24TECHREAD_AUTH_USERNAME",
    "W24TECHREAD_AUTH_PASSWORD",
    "W24TECHREAD_AUTH_REGION"
]
""" List of the environment variables used by the
client """


class Hook(BaseModel):
    """ Small Object to keep the callback requests.
    You can either register a callback request to an
    ask or a message_type.

    If you register an ask, be sure to use a complete W24Ask
    definition; not just the ask type.
    """

    message_type: Optional[W24TechreadMessageType]
    message_subtype: Optional[W24TechreadMessageSubtype]
    ask: Optional[W24Ask]
    function: Callable


class W24TechreadClient:
    """ Simple W24Client that allows you to use
    learn more about the content on your Technical
    Drawings.
    """

    def __init__(
            self,
            techread_server_https: str,
            techread_server_wss: str,
            techread_version: str,
            development_key: str = None):
        """ Initialize a new W24TechreadClient. If you wonder
        about any of the attributes, have a look at the .env
        file that we provided to you. They contain all the
        information that you will need.

        Arguments:
            techread_server_https {str} -- domain name that
                is being used by the https client

            techread_server_wss {str} -- domain name that
                is being used by the websocket client

            techread_version {str} -- version that you want to
                connect to

            development_key {str} -- key that allows you to submit
                your request to one of the internal architectures.
                You can try guessing or bruteforcing this key;
                we'll just charge you for every request you submit and
                transfer the money to the holiday bonus account.
        """

        # save the development_key
        self._development_key = development_key

        # Create an empty reference to the authentication
        # service (currently AWS Cognito)
        self._auth_client: Optional[AuthClient] = None

        # Initialize an instance of the HTTPS client
        self._techread_client_https = TechreadClientHttps(
            techread_server_https, techread_version)

        # Initialize an instance of the WEBSCOKET client
        self._techread_client_wss = TechreadClientWss(
            techread_server_wss, techread_version)

    async def __aenter__(
            self
    )-> 'W24TechreadClient':
        """ Create the HTTPS and WSS sessions

        Raises:
            RuntimeError: Exception is raised if
                you tried to enter a session before
                calling the register() method

        Returns:
            W24TechreadClient -- Version of self with
                active sessions
        """

        # ensure that we have a token
        try:
            await self._auth_client.login()  # type: ignore
        except AttributeError:
            raise RuntimeError(
                "No connection to the authentication service was " +
                "established. Please call register()")

        # enter the https session
        await self._techread_client_https.__aenter__()

        # enter the wss session
        await self._techread_client_wss.__aenter__()

        # return the "entered" version of self
        return self

    async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_value: Optional[BaseException],
            traceback: Optional[TracebackType]
    ) -> None:

        """ Ensure that the sessions are closed
        """

        # close the HTTPS session
        await self._techread_client_https.__aexit__(
            exc_type, exc_value, traceback)

        # close the WSS session
        await self._techread_client_wss.__aexit__(
            exc_type, exc_value, traceback)

    def register(
            self,
            cognito_region: str,
            cognito_identity_pool_id: str,
            cognito_user_pool_id: str,
            cognito_client_id: str,
            cognito_client_secret: str,
            username: str,
            password: str
    ) -> None:
        """
        Register with the authentication
        service (i.e., lazy login)

        Arguments:
            cognito_region {str} -- Physical region
            cognito_identity_pool_id {str} -- identity pool of W24
            cognito_client_id {str} -- the client id of your application
            cognito_client_secret {str} -- the client secrect of your
                application
            username {str} -- the username with which you want to register
            password {str} -- the password with which you want to register
        """
        # create an client instance to connect
        # to the authentication service
        self._auth_client = AuthClient(
            cognito_region,
            cognito_identity_pool_id,
            cognito_user_pool_id,
            cognito_client_id,
            cognito_client_secret)

        # register username and password
        self._auth_client.register(username, password)

        # tell the techread clients about it
        self._techread_client_https.register_auth_client(self._auth_client)
        self._techread_client_wss.register_auth_client(self._auth_client)

    @property
    def username(self) -> Optional[str]:
        """ Make the username accessable to the CLI and GUI

        Returns:
            str: username of the currently registered user
        """
        try:
            return self._auth_client.username
        except ValueError:
            return None

    async def read_drawing(
        self,
        drawing: bytes,
        asks: List[W24Ask],
        model: bytes = None
    ) -> AsyncGenerator:
        """ Send a Technical Drawing to the W24 API to have it automatically
        interpreted and read. The API will return

        Arguments:
            drawing {bytes} -- binary representation of a technical drawing.
                Please refer to the API - documentation to learn which mime
                types are currently supported

        Keyword Arguments:
            model {bytes} -- binary represetation of the 3d model (typically
                step)
                Please refer to the API - documentation to learn whcih mime
                types are currently sypported(default: {None})

            asks {List[W24Ask]} --
                List of Asks that are requested from the API. They must derive
                from the W24Ask object. Refer to the API documentation for
                a full list of supported W24AskTypes

        Yields:
            W24TechreadMessage -- Response object obtained from the API
                that indicates the state of your request. Be sure to pass this
                to the read_drawing_listen method

        Raises:
            DrawingTooLarge -- Exception is raised when the drawing was too
                large to be processed. At the time of writing. The upload
                limit lies at 6 MB (including overhead).
        """

        # give us some debug information
        logger.info("API method read_drawing() called")

        # tell us when a development key is being used
        if self._development_key:  # pragma: no cover
            logger.info("Using development key %s***",
                        self._development_key[:8])

        # make the request object
        request = W24TechreadRequest(
            asks=asks,
            development_key=self._development_key)

        # send the initialization request to the server.
        # This achieves two things:
        # 1. The server has a couple of 100ms to
        #    reserves some resources for you, and
        # 2. The server will create a new request_id
        #    that you will need when uploading the
        #    associated files
        await self._techread_client_wss.send_command(
            W24TechreadAction.INITIALIZE.value,
            request.json())

        # Wait for the response (i.e,. the request id)
        response = await self._techread_client_wss.recv_message()
        logger.info("Received request_id %s", response.request_id)

        # interpret the payload
        init_response = W24TechreadInitResponse.parse_obj(
            response.payload_dict)

        # upload drawing and model. We can do that in parallel.
        # If your user uploads them separately, you could also
        # upload them separately to Werk24.
        try:
            await asyncio.gather(
                self._techread_client_https.upload_associated_file(
                    init_response.drawing_presigned_post,
                    drawing),
                self._techread_client_https.upload_associated_file(
                    init_response.model_presigned_post,
                    model))

        # explicitly reraise the exception if the payload is too
        # large
        except RequestTooLargeException:  # pylint: disable=try-except-raise
            raise

        # Tell Werk24 that all the files have been uploaded
        # correctly and the reading process can be started.
        #
        # NOTE: you will only be able to start the reading
        # process from the websocket connection that
        # initiated the request. If you want to run a
        # stateless-system that separates the initialization
        # from the upload and read stages, you'll need to
        # find a way of handing over the tcp connection :)
        # PS: The AWS API Gatway for websockets might help you
        # here.
        await self._techread_client_wss.send_command(
            W24TechreadAction.READ.value,
            "{}")
        logger.info("Techread request submitted")

        # Wait for incoming messages from the server.
        # They will tell you when the individual
        # asks become available. The socket returns
        # strings of jsonified W24TechreadMessage objects.
        #
        # The loop will stop when the websocket is
        # closed
        async for message in self._techread_client_wss.listen():

            # check whether we need to download something
            if message.payload_url is not None:
                message.payload_bytes = await self._techread_client_https \
                    .download_payload(message.payload_url)

            # return the message to the caller for immediate
            # consumption
            yield message

    @staticmethod
    def make_from_env(
        license_path: Optional[str] = ".werk24"
    ) -> "W24TechreadClient":
        """ Small helper function that creates a new
        W24TechreadClient from the enviorment info.

        Arguments:
            license_path:{Optional[str]} -- path to the License file.
                By default we are looking for a .werk24 file in the current
                cwd. If argument is set to None, we are not loading any
                file and relying on the ENVIRONMENT variables only

        Raises:
            FileNotFoundError -- Raised when you pass a path to a license file
                that does not exist
            UnauthorizedException -- Raised when the credentials were not
                accepted by the API

        Returns:
            W24TechreadClient -- The techread Client
        """

        # First priority: look for the local license path
        if license_path is not None:
            if os.path.exists(license_path):
                environ_raw = {
                    k: v
                    for k, v in dotenv.dotenv_values(license_path).items()
                    if v is not None}
            else:
                environ_raw = {}

        # Second priority: use the environment variables
        else:
            environ_raw = dict(os.environ)

        # get the variables from the environment and ensure that they
        # are set. If not, raise an exception
        try:
            environs = {cur_key: environ_raw[cur_key]
                        for cur_key in ENVIRONS}

            # create a reference to the client
            client = W24TechreadClient(
                environs['W24TECHREAD_SERVER_HTTPS'],
                environs['W24TECHREAD_SERVER_WSS'],
                environs['W24TECHREAD_VERSION'])

            # register the credentials. This will in effect
            # only set the variabels in the authorizer. It will
            # not trigger a network request
            client.register(
                environs['W24TECHREAD_AUTH_REGION'],
                environs['W24TECHREAD_AUTH_IDENTITY_POOL_ID'],
                environs['W24TECHREAD_AUTH_USER_POOL_ID'],
                environs['W24TECHREAD_AUTH_CLIENT_ID'],
                environs['W24TECHREAD_AUTH_CLIENT_SECRET'],
                environs['W24TECHREAD_AUTH_USERNAME'],
                environs['W24TECHREAD_AUTH_PASSWORD'])

        except KeyError:
            raise LicenseError(
                "The License information could neither be "
                "found in the local environment variables, nor in the "
                "local '.werk24' file. Please make sure that you are "
                "calling the client from the directory that contains "
                "your '.werk24' file and that the license file "
                "name does not contain a prefix.")

        # return the client
        return client

    async def read_drawing_with_hooks(
        self,
        drawing_bytes: bytes,
        hooks: List[Hook]
    ) -> None:
        """ Send the drawing to the API (can be PDF or image)
        and register a number of callbacks that are triggered
        once the asks become available.

        Arguments:
            drawing_bytes {bytes} -- Technical Drawing as Image or PDF
            hooks {List[Hook]} -- List of Callback you want to obtain

        Raises:
            ServerException -- Raised when the server returns an ERROR
                message
        """

        # filter the callback requests to only contain
        # the ask types
        asks_list = [
            cur_ask.ask
            for cur_ask in hooks
            if cur_ask.ask is not None]

        try:
            # send out the request and make a generator
            # that triggers when the result of an ask
            # becomes available
            async for message in self.read_drawing(drawing_bytes, asks_list):
                await self._call_hooks_for_message(message, hooks)

        # explicitly reraise server exceptions
        except ServerException:  # pylint: disable=try-except-raise
            raise

        # explicitly reraise RequestTooLargeException
        except RequestTooLargeException:  # pylint: disable=try-except-raise
            raise

    async def _call_hooks_for_message(
            self,
            message: W24TechreadMessage,
            hooks: List[Hook]
    ) -> None:
        """ Find the correct hook for the read reseponse and
        call the corresponding hook.

        Arguments:
            message {W24TechreadMessage} -- Messsage returned from the
                read_drawing method

            hooks {List[Hook]} -- List of hooks from which we need to
                pick the suited one

        Raises:
            ServerException: raised when the server returns an ERROR
                message
        """

        # get the hook function that corresponds to the message
        hook_function = self._get_hook_function_for_message(message, hooks)

        # if no hook is registered, ignore
        if hook_function is None:
            return

        # if the hook_function is not callable, we want to warn the user,
        # rather than throwing an exception
        if not callable(hook_function):
            logger.warning(
                "You registered a non-callable trigger of type '%s' with "
                "the message_type '%s'. Please make sure that you are using "
                "a Callable (e.g, def or lambda)",
                type(hook_function),
                message.message_type)
            return

        # if everything went well, we call the trigger with
        # the message as payload. Be sure to call the
        # function asymmetrically if supported
        if asyncio.iscoroutinefunction(hook_function):
            await hook_function(message)
        else:
            hook_function(message)

    @staticmethod
    def _get_hook_function_for_message(
            message: W24TechreadMessage,
            hooks: List[Hook]
    ) -> Optional[Callable]:
        """ Get the hook function that corresponds to the message
        type.

        Arguments:
            message {W24TechreadMessage} -- Messsage returned from the
                read_drawing method

            hooks {List[Hook]} -- List of hooks from which we need to
                pick the suited one

        Returns:
            Optional[Callable] -- Hook function that should be called
        """

        # because we allow the user to define the ask in the hook,
        # we need to make some extra effort when filtering
        if message.message_type == W24TechreadMessageType.ASK:
            def hook_filter(hook: Hook) -> bool:
                return hook.ask is not None \
                    and message.message_subtype.value \
                    == hook.ask.ask_type.value

        # if the message is of any other type, we just need to
        # compare the the message_type and message_subtype
        else:
            def hook_filter(hook: Hook) -> bool:
                return hook.message_type is not None \
                    and hook.message_subtype is not None \
                    and message.message_type == hook.message_type \
                    and message.message_subtype == hook.message_subtype

        # return the first positive case
        for cur_hook in filter(hook_filter, hooks):
            return cur_hook.function

        # if we are still here, we have an unknown message type, which
        # probobly is being caused by an API update. We want to ensure
        # that the user is being informed, but we do not want to break
        # the existing functionality -> warning
        logger.warning(
            "Ignoring message of type %s:%s - no hook registered",  # noqa
            message.message_type,
            message.message_subtype)
        return None



""" Definition of the Thread support structures

Author: Jochen Mattes - Werk24
"""
from enum import Enum
from pydantic import BaseModel
from typing import Optional
import abc


class W24ThreadType(str, Enum):
    """ Enum for the individual thread types
    """
    ISO_METRIC = "ISO_METRIC"
    WHITWORTH = "WHITWORTH"
    UTS_COARSE = "UTS_COARSE"
    UTS_FINE = "UTS_FINE"
    UTS_EXTRAFINE = "UTS_EXTRAFINE"
    UTS_SPECIAL = "UTS_SPECIAL"


class W24Thread(BaseModel, abc.ABC):
    """ Abstract Base Class for all Threads
    """

    thread_type: W24ThreadType
    """ thread type to facilitate deserialization
    """

    blurb: str
    """ string representation for human interpretation
    """


class W24ThreadHandedness(str, Enum):
    """ Enum describing the direction of the
    thread
    """
    LEFT = "LEFT"
    RIGHT = "RIGHT"


class W24ThreadISOMetric(W24Thread):
    """ Metric ISO Thread following ISO 1502

    Supports
    * DIN 14-1 threads (i.e., diameter < 1mm)
    * DIN 13-1 threads (i.e., diameter 1..68mm)
    * DIN 13-2 to DIN 13-10 threads (i.e., diameter 1..1000mm)
    * DIN 158-1 (i.e., cone-shaped threads)
    """
    thread_type = W24ThreadType.ISO_METRIC

    diameter: float
    """ Diameter of the thread in mm.

    NOTE: The norms range from 0.1 to 1000mm;
    however, diameters outside that range are possible
    (and occur)
    """

    pitch: Optional[float] = None
    """ Pitch of the thread in mm.
    Normed range: 0.25 - 9mm

    NOTE: the value will only be set if is explicitly
    stated in the drawing. This behaviour might change
    in the future and perform an automatic lookup in the
    DIN standards. If this change would affect your
    application, please get in touch with us.
    """

    # shape: W24ThreadMetricISOShape
    """ NOTE: cone-shaped threads identified by a 'keg' suffix
    are currently not supported.
    """  # pylint: disable=pointless-string-statement

    handedness: W24ThreadHandedness = W24ThreadHandedness.RIGHT
    """ Left of right-handedness of the thread.
    This will be RIGHT unless explicitly described as LEFT
    in the drawing
    """


class W24ThreadUTS(W24Thread):
    """ Unified Thread Standard (UTS) base class for
    * UNC - Unified National Coarse Thread
    * UNF - Unified National Fine Thread
    * UNEF - Unified National Extrafine Thread
    """

    uts_size: str
    """ UTS size as string representation.
    Threads with a diameter < 0.25 inch are written with a leading '#'.
    Threads with a diameter >= 0.25 inch are represented as fractions
        with a tailing '"'
    Examples: #0, 1 3/4"
    """

    diameter: float
    """ diameter in inch derived from the unc_size
    """

    threads_per_inch: float
    """ Threads per inch
    NOTE: The float (rather than int) is chosen to support non-convertional
    threads as well.
    """

    tolerance_class: str
    """ Tolerance class
    Options:
    * 1A, 2A, 3A for external threads
    * 1B, 2B, 3B for internal threads
    """


class W24ThreadUTSCoarse(W24ThreadUTS):
    """ Unified National Coarse Thread
    """
    thread_type = W24ThreadType.UTS_COARSE


class W24ThreadUTSFine(W24ThreadUTS):
    """ Unified National Fine Thread
    """
    thread_type = W24ThreadType.UTS_FINE


class W24ThreadUTSExtrafine(W24ThreadUTS):
    """ Unified National Extrafine Thread
    """
    thread_type = W24ThreadType.UTS_EXTRAFINE


class W24ThreadUTSSpecial(W24ThreadUTS):
    """ Unified National Special Thread
    """
    thread_type = W24ThreadType.UTS_SPECIAL


class W24ThreadWhitworth(W24Thread):
    """ Whitworth Thread following ISO 228-1
    """
    thread_type = W24ThreadType.WHITWORTH

    whitworth_size: float
    """ Size number (for historic reasons not proportional to size)
    """

    tolerance_class: Optional[str] = None
    """ Tolerance field A implemented in the future
    """  # pylint: disable=pointless-string-statement



""" Command Line Interface for W24 Techread
"""
import argparse
import io
import logging
import sys
from collections import namedtuple
from typing import List

from dotenv import load_dotenv
from werk24.exceptions import RequestTooLargeException
from werk24.models.ask import (W24AskCanvasThumbnail, W24AskPageThumbnail,
                               W24AskSectionalThumbnail, W24AskSheetThumbnail,
                               W24AskVariantCAD, W24AskVariantGDTs,
                               W24AskVariantLeaders, W24AskVariantMeasures)
from werk24.models.techread import (W24TechreadMessageSubtypeError,
                                    W24TechreadMessageSubtypeProgress,
                                    W24TechreadMessageType)
from werk24.techread_client import Hook, LicenseError, W24TechreadClient

# load the environment variables
load_dotenv(".werk24")

# set the log level to info for the test setting
# We recommend using logging.WARNING for production
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s',  # noqa
    datefmt='%Y-%m-%d %H:%M:%S')

# get the local logger
logger = logging.getLogger(__name__)  # pylint:disable = invalid-name

# make the configuration of what hooks we want to handle and how
HookConfig = namedtuple("HookkConfig", "arg ask function")
hook_config = [
    HookConfig(
        'ask_page_thumbnail',
        W24AskPageThumbnail,
        lambda m: _show_image("Page Thumbnail", m.payload_bytes)),
    HookConfig(
        'ask_sheet_thumbnail',
        W24AskSheetThumbnail,
        lambda m: _show_image("Sheet Thumbnail", m.payload_bytes)),
    HookConfig(
        'ask_canvas_thumbnail',
        W24AskCanvasThumbnail,
        lambda m: _show_image("Canvas Thumbnail", m.payload_bytes)),
    HookConfig(
        'ask_sectional_thumbnail',
        W24AskSectionalThumbnail,
        lambda m: _show_image("Sectional Thumbnail", m.payload_bytes)),
    HookConfig(
        'ask_variant_measures',
        W24AskVariantMeasures,
        lambda m: print("Ask Variant Measures\n", m.payload_dict)),
    HookConfig(
        'ask_variant_measures',
        W24AskVariantMeasures,
        lambda m: print("Ask Variant Measures\n", m.payload_dict)),
    HookConfig(
        'ask_variant_gdts',
        W24AskVariantGDTs,
        lambda m: print("Ask Variant GDTs\n", m.payload_dict)),
    HookConfig(
        'ask_variant_leaders',
        W24AskVariantLeaders,
        lambda m: print("Ask Variant Leaders\n", m.payload_dict)),
    HookConfig(
        'ask_variant_cad',
        W24AskVariantCAD,
        lambda m: print("Ask Variant Cad\n", m.payload_dict)),
]


def _get_drawing(
        file_path: str
) -> bytes:
    """ Get the bytes of the file that shall be
    read.

    Args:
        file_path (str): path to the input file

    Returns:
        bytes: content of the fiel
    """
    # get the content
    with open(file_path, "rb") as filehandle:
        return filehandle.read()


def _show_image(
        log_text: str,
        image_bytes: bytes
) -> None:
    """ Display a debug image. The function relies on
    the PIL functionality of Image.show(). The
    actual behaviour will thus be diffent accross platforms

    Args:
        log_text (str): Log Text that we want to write back
            to the command line interface
        image_bytes (bytes): byte representation of the image
            that is to be displayed
    """

    # Import the PIL at runtime to make sure that
    # you can still use the client without the need
    # to install Pillow
    try:
        from PIL import Image  # pylint: disable=import-outside-toplevel
    except ImportError:
        logger.warning(
            "Viewing image-like responses requires the installation"
            " of the PIL package: pip install pillow. Image skipped.")
        return

    # otherwise print the text
    logger.info(log_text)

    # and show the image
    image = Image.open(io.BytesIO(image_bytes))
    image.show(title=log_text)


async def main(
    args: argparse.Namespace
) -> None:
    """
    Run the Techread command with the CLI arguments

    Args:
        args(argparse.Namespace): CLI args generated by
        argparse
    """

    # make the hooks from the arguments
    hooks = _make_hooks_from_args(args)

    # make the client. This will automatically
    # fetch the authentication information
    # from the environment variables. We will
    # provide you with separate .env files for
    # the development and production environments
    try:
        client = W24TechreadClient.make_from_env()

    # If a license error occured, let the user know.
    # NOTE: This will not catch deactivated users.
    # They will only learn about their status once
    # they send a request.
    except LicenseError as exception:
        print(f"LICENSE ERROR: {exception}")
        sys.exit()

    async with client as session:

        # get the drawing
        drawing_bytes = _get_drawing(args.input_file)

        # and make the request
        try:
            await session.read_drawing_with_hooks(
                drawing_bytes,
                hooks)

        except RequestTooLargeException:
            message = "Request was too large to be processed. " \
                + "Please check the documentation for current limits."
            logger.error(message)


def _make_hooks_from_args(
        args: argparse.Namespace
) -> List[Hook]:

    # add a general hook to deal with internal errors
    hook_error_internal = Hook(
        message_type=W24TechreadMessageType.ERROR,
        message_subtype=W24TechreadMessageSubtypeError.INTERNAL,
        function=lambda msg:
        logger.error("Internal Error %s", msg.payload_dict))

    # add a gneral hook to deal with PROGRESS messages
    hook_progress_started = Hook(
        message_type=W24TechreadMessageType.PROGRESS,
        message_subtype=W24TechreadMessageSubtypeProgress.STARTED,
        function=lambda msg: logger.info("Techread process started"))

    # tell the api what asks you are interested in,
    # and define what to do when you receive the result
    hooks = [
        Hook(ask=c.ask(), function=c.function)
        for c in hook_config
        if getattr(args, c.arg)] + \
        [
        hook_error_internal,
        hook_progress_started]

    return hooks



""" Defintion of all W24Ask types that are understood by the Werk24 API.
"""
from enum import Enum
from typing import List, Optional, Union

from pydantic import UUID4, BaseModel

from .file_format import FileFormatCad, FileFormatImage
from .gdt import W24GDT
from .leader import W24Leader
from .measure import W24Measure


class W24AskType(str, Enum):
    """ List of all Ask Type supported by the current
    API version.

    """
    CANVAS_THUMBNAIL = "CANVAS_THUMBNAIL"
    """ Thumbnail of the canvas (i.e., the part of the
    sheet that contains the geometry)
    """

    PAGE_THUMBNAIL = "PAGE_THUMBNAIL"
    """ Thumbnail of the overall page - rotated and with
    surrounding white space removed
    """

    SECTIONAL_THUMBNAIL = "SECTIONAL_THUMBNAIL"
    """ Thumbnail of a sectional on the canvas.
    Here the sectional describes both cuts and perspectives
    """

    SHEET_THUMBNAIL = "SHEET_THUMBNAIL"
    """ Thumbnail of the sheet (i.e., the part of the
    page that is described by the surrounding frame)
    """

    TRAIN = "TRAIN"
    """ Supplying the request for training only without
    expecting a response.
    """

    VARIANT_CAD = "VARIANT_CAD"
    """ Requests the generation of a CAD file
    """

    VARIANT_GDTS = "VARIANT_GDTS"
    """ List of Geometric Dimensions and Tolerations detected
    on the Sectionals associated with the variant
    """

    VARIANT_LEADERS = "VARIANT_LEADERS"
    """ List of Leaders that were detected on the Sectional
    """

    VARIANT_MATERIAL = "VARIANT_MATERIAL"
    """ Material that was detected on the data fields of the
    drawing or within a variant table
    """

    VARIANT_MEASURES = "VARIANT_MEASURES"
    """ List of Measures that were found on the Sectionals
    associated with the variant
    """


class W24Ask(BaseModel):
    """ Base model from wich all Asks inherit
    """
    ask_type: W24AskType

    is_training: bool = False
    """ Flag that indicates that your request is a pure
    training request and that you are not expecting to
    obtain a response.
    """


class W24AskThumbnail(W24Ask):
    """ Base model for features that request a thumbnail.

    NOTE: At this stage, the API will return a high-resolution
    image. Future releases will allow the defintion of the
    maximal dimensions.
    """
    file_format: FileFormatImage = FileFormatImage.JPEG


class W24AskPageThumbnail(W24AskThumbnail):
    """ Requests a thumbnail for each page in the document;
    rotated, and with the surrounding white-space removed.

    NOTE: when you supply a white-on-black document, the thumbnail
    will be black-on-white.
    """
    ask_type = W24AskType.PAGE_THUMBNAIL


class W24AskSheetThumbnail(W24AskThumbnail):
    """ Requests a thumbnail of each sheet on each page in
    the document. The sheet will only contain the pixels within
    the main frame that surrounds the canvas and header fields.
    """
    ask_type = W24AskType.SHEET_THUMBNAIL


class W24AskCanvasThumbnail(W24AskThumbnail):
    """ Requests a thumbnail of each canvas in each sheet.
    The canvas describes the "drawing area" of the sheet.
    """
    ask_type = W24AskType.CANVAS_THUMBNAIL


class W24AskSectionalThumbnail(W24AskThumbnail):
    """ The W24AskPlaneThumbnail requests a thumbnail
    of each sectional on each sheet in the document.

    NOTE: we have chosen the term sectional to describe
    both cuts and perspectives
    """
    ask_type = W24AskType.SECTIONAL_THUMBNAIL


class W24AskVariantMeasures(W24Ask):
    """ With this Ask you are requesting the complete
    list of all measures that were detected for the
    variant
    """
    ask_type = W24AskType.VARIANT_MEASURES

    confidence_min: float = 0.2
    """ Werk24 calculates internal confidence scores
    for each measure. Depending on your use-case you
    might want to consider or discard low-confidence
    results. This parameter allows you to filter the
    results. The resulting W24Measure objects also
    contain a confidence score that allows you to filter
    even further.
    """


class W24AskVariantMeasuresResponse(BaseModel):
    """ Response object corresponding to the
    W24AskVariantMeasures ask.

    NOTE: Be aware that requesting the measures will
    yield one responds for each variant and sectional
    """
    variant_id: UUID4
    sectional_id: UUID4
    measures: List[W24Measure]


class W24AskVariantLeaders(W24Ask):
    """ With this Ask you are requesting the complete
    list of all leaders that were detected on the
    variant
    """
    ask_type = W24AskType.VARIANT_LEADERS


class W24AskVariantLeadersResponse(BaseModel):
    """ Response object corresponding to the
    W24AskVariantLeaders.

    """
    variant_id: UUID4
    sectional_id: UUID4
    leaders: List[W24Leader]


class W24AskVariantMaterial(W24Ask):
    """ This ask requests the material of the
    individual variant.

    !!! This Ask will not be answered by
    !!! the API at this stage. Stay tuned
    !!! for updates
    """
    ask_type = W24AskType.VARIANT_MATERIAL

    material_hint: Optional[str] = None
    """ If your user is already asked about the
    meterial, or if you have any additional
    informationa about the material, you
    can submit it here. This will make the
    reading more stable.
    """


class W24AskVariantGDTs(W24Ask):
    """ This Ask requests the complete
    list of all Geometric Dimensions and Tolerations
    that were detected on the variant.
    """
    ask_type = W24AskType.VARIANT_GDTS


class W24AskVariantGDTsResponse(BaseModel):
    """ Response object corresponding ot the
    W24AskVariantGDTs.

    NOTE: Be aware that requesting the measures will
    yield one response for each variant and sectional
    """
    variant_id: UUID4
    sectional_id: UUID4
    gdts: List[W24GDT]


class W24AskTrain(W24Ask):
    """ If you submit this Ask, we will use your request
    to train and improve our models.
    It does not trigger a response
    """
    ask_type = W24AskType.TRAIN


class W24AskVariantCAD(W24Ask):
    """ By sending this ASk, you are requesting
    an associated CAD model
    """
    ask_type = W24AskType.VARIANT_CAD

    output_format: FileFormatCad = FileFormatCad.DXF
    """ Output format in which to generate
    the CAD file
    """


class W24AskVariantCADResponse(BaseModel):
    """ Response object corresponding ot the
    W24AskVariantCad.

    NOTE: Be aware that requesting the measures will
    yield one response for each variant

    NOTE: the cad file will be returned as part of
    the payload_bytes and needs to be accessed directly
    """
    variant_id: UUID4


W24AskUnion = Union[
    W24AskCanvasThumbnail,
    W24AskPageThumbnail,
    W24AskSectionalThumbnail,
    W24AskSheetThumbnail,
    W24AskTrain,
    W24AskVariantGDTs,
    W24AskVariantLeaders,
    W24AskVariantMaterial,
    W24AskVariantMeasures,
]
""" Union of all W24Asks to ensure proper deserialization """




""" Defintion of all the W24Measure class its support structures


Author: Jochen Mattes - Werk24
"""
from typing import List, Optional, Tuple

from pydantic import BaseModel, UUID4

from .measure_warning import W24MeasureWarning
from .thread import W24Thread
from .unit import W24UnitLength
from .size_tolerance import W24SizeTolerance, W24SizeToleranceGeneral
from .chamfer import W24Chamfer
from .size import W24Size


class W24MeasureLabel(BaseModel):
    """ Abstract base class for all MeasureItem Children
    """

    blurb: str
    """ String representation of the item for human consumption
    """

    quantity: int = 1
    """ Quantity for spacings
    e.g., (2x)

    NOTE: Currently only the indicated measure will be detected and
    returned. Future implementation could detect the spacings and
    return individual measures for the respective spacings.
    When we implement this feature, we will add an attribute to the
    ASK, which allows you to control the behavior.
    """

    size: W24Size
    """ Size of measure as refered to in the drawing.

    NOTE: Sizes are always assocated with units! This becomes important
    to remember when you are dealing with mixed-unit drawings (e.g.,
    an adapter bolt that has an ISO thread and an UTC tapped hole).
    To avoid any loss of precision, we return the size in the unit
    that was indicated on the drawing.
    """

    unit: Optional[W24UnitLength] = None
    """ Length unit of the size
    """

    size_tolerance: W24SizeTolerance = W24SizeToleranceGeneral()
    """ Tolerance details.
    Default: General tolerances

    NOTE: by default we are refering to the general tolerances of the drawing.
    Currently the W24SizeToleranceGeneral object is a stub.
    Future implementations might go one step further and quote the
    applicable general tolerance as refered to in the data fields.

    NOTE: if the W24MeasureLabel describes a "Theoretically Exact Measure",
    i.e, the label is surrounded by a box, like: "[15]", the size_tolerance
    refers to a W24SizeToleranceTheoreticallyExact object (and is NOT None)
    """

    thread: Optional[W24Thread] = None
    """ Optional thread details.

    NOTE: the thread details describe the complete thread description
    and follow the respective standards. In consequence, the thread
    diameter of an ISO-thread will be indicated in millimeter,
    while the thread diameter of an UTS thread will be in inch.

    """

    chamfer: Optional[W24Chamfer] = None
    """ Optional Chamfer
    """


class W24Measure(BaseModel):
    """ Tolerated measure
    """

    measure_id: Optional[UUID4] = None
    """ Unique id of measure
    """

    line: Tuple[Tuple[float, float], Tuple[float, float]]
    """ Tuple of the measure's start- and end-coordinates
    in the Pixel Coordinate system of the sectional.

    The coordinates are normalized by the width and height of the
    associated object (e.g., the sectional). If you want to obtain
    the absolute position in the original image, you need to consider
    the following offsets: sectional + canvas + sheet.
    """

    label: W24MeasureLabel
    """ Measure Label
    """

    warnings: List[W24MeasureWarning] = []
    """ List of Warnings that are associated with the
    measure. See W24MeasureWarning for details
    """

    confidence: float = 0.0
    """ Werk24 calculates an internal confidence score for
    each measure. Depending on your use-case, you might want
    to consider or discard low-confidence measures. This
    value allows you to do so. The value ranges from
    0.0 to 1.0
    """



""" Defintion of the W24Chamfer class and its support structures


Author: Jochen Mattes - Werk24
"""
from pydantic import BaseModel


class W24Chamfer(BaseModel):
    """ Chamfer in degree.

    NOTE: The chamfers can be tolerated.
    Future implementations wil take this into account.
    """

    blurb: str
    """ String representation for human consumption
    """

    angle: float
    """ Chamfer angle in degrees

    NOTE: If you are dealing with GON, let us know.
    Happy to do exend the API for the pleasure of
    dealing with the concept.
    """



import abc
from enum import Enum

from pydantic import BaseModel


class W24SizeToleranceType(str, Enum):
    """ Enum of the supported tolerances
    """
    APPROXIMATION = "APPROXIMATION"
    FIT_SIZE_ISO = "FIT_SIZE_ISO"
    GENERAL_TOLERANCES = "GENERAL_TOLERANCES"
    MINIMUM = "MINIMUM"
    MAXIMUM = "MAXIUM"
    OFF_SIZE = "OFF_SIZE"
    THEORETICALLY_EXACT = "THEORETICALLY_EXACT"


class W24SizeTolerance(BaseModel, abc.ABC):
    """ Abstract Base Class to cover the Tolerations
    """

    toleration_type: W24SizeToleranceType
    """ Toleration Type for deserialization
    """

    blurb: str
    """ String representation for human consumption
    """


class W24SizeToleranceFitsizeISO(W24SizeTolerance):
    """ ISO fit size tolerations
    """
    toleration_type = W24SizeToleranceType.FIT_SIZE_ISO

    blurb: str
    """ uninterpreted string representation of the fit size
    """


class W24SizeToleranceOffSize(W24SizeTolerance):
    """ Off-size based tolerations
    """
    toleration_type = W24SizeToleranceType.OFF_SIZE

    deviation_lower: float
    """ Lower deviation from the nominal size
    """

    deviation_upper: float
    """ Upper deviation from the nominal size
    """


class W24SizeToleranceGeneral(W24SizeTolerance):
    """ General Tolerances
    """
    toleration_type = W24SizeToleranceType.GENERAL_TOLERANCES

    blurb = ""
    """ Set the blub to '' (default for General Tolerances)
    """


class W24SizeToleranceTheoreticallyExact(W24SizeTolerance):
    """ Theoretically Exact Measures after ISO 5458
    must not be tolerated. They are indicated by a small
    rectangular frame.

    Example:
        +------+
        |  15  |
        +------+

    NOTE: In practice, we see Technical drawings contain
    tolerated, theoretically exact measures.
    Example:
        +------------+
        | 15 +/- 0.2 |
        +------------+

    In these situations the toleration takes priority.
    """
    toleration_type = W24SizeToleranceType.THEORETICALLY_EXACT


class W24SizeToleranceMinimum(W24SizeTolerance):
    """ Minimum Size of a measure

    Example:
        min. 15
    """

    toleration_type = W24SizeToleranceType.MINIMUM


class W24SizeToleranceMaximum(W24SizeTolerance):
    """ Maximum Size of a measure

    Example:
        max 15
    """

    toleration_type = W24SizeToleranceType.MINIMUM


class W24SizeToleranceApproximation(W24SizeTolerance):
    """ Approximation of a measure

    Example:
        ~ 15
        ca. 14
    """
    toleration_type = W24SizeToleranceType.APPROXIMATION
